#include <string>
#include <iostream>
#include <map>
#include <vector>
#include <sstream>


#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

namespace LTF {
	class Type;

	class Tokenizer;
	class Symbol;
	class Program;
	class Type;
	class TypeBits;
	class TypeFBits;
	class TypeArrayAST;
	class VariableAST;
	class TypeVariableAST;
	class ValueAST;
	class IntValueAST;
	class StringValueAST;
	class VariableValueAST;
	class CodeAST;
	class NewCodeAST;
	class FunctionAST;
	class OutputFunctionAST;
	class InputFunctionAST;

	class Type {
		private:
			bool Inline;
			bool NoRef;
			bool NoLock;
			Type *parent;
			std::map<std::string, TypeVariableAST *> *Members;
			std::vector<FunctionAST *> Functions;
		public:
			std::string *Name;
			explicit Type(std::string *name) : Name(name) , Inline(false) , NoRef(false), NoLock(false) , parent(NULL) { Members = new std::map<std::string, TypeVariableAST *>; }
			bool setInline();
			bool isInline();
			bool setNoRef();
			bool setNoLock();
			bool setParent(Type *p);
			bool addMember(TypeVariableAST *v);
			bool addFunction(FunctionAST *f);
			int countArguments();
	};
	
	class TypeBits : public Type {
		private:
			bool Signed;
		public:
			explicit TypeBits(bool Signed) : Signed(Signed) , Type(new std::string("bits")) {};
	};
	
	class TypeFBits : public Type {
		private:
		public:
			explicit TypeFBits() : Type(new std::string("fbits")) {};
	};
	
	class TypeArrayAST {
		private:
			Type *t;
			std::vector<unsigned long long> ArraySize;
		public:
			TypeArrayAST(Type *t) : t(t) {}
			void addLevel(int size);
			int getArrayDepth();
			int getSizeAtLevel(int level);
		
	};
	
	class VariableAST {
		private:
			TypeArrayAST *t;
			std::string *name;
		public:
			explicit VariableAST(std::string *name, TypeArrayAST *t) : name(name) , t(t) {}
			std::string *getName() {return this->name; }
	};
	
	class TypeVariableAST : public VariableAST {
		private:
			bool init;
		public:
			explicit TypeVariableAST(std::string *name, TypeArrayAST *t) : init(false) , VariableAST(name, t) {}
			bool setInit() { if(this->init) return false; this->init = true; return true; };
			bool getInit() {return this->init; };
	};
	
	class ValueAST {
		private:
			std::string *name;
		public:
			explicit ValueAST(std::string *name) : name(name) {}
			std::string *getName() {return this->name; }
	};
	
	class IntValueAST : public ValueAST {
		private:
			unsigned long long value;
			bool negative;
		public:
			explicit IntValueAST(unsigned long long value, bool neg) : value(value), negative(neg) , ValueAST(new std::string("Integer")) {}
			unsigned long long getValue() { return this->value; };
			bool isNeg() { return this->negative; };
	};
	
	class StringValueAST : public ValueAST {
		private:
			std::string *value;
		public:
			explicit StringValueAST(std::string *value) : value(value), ValueAST(new std::string("String")) {}
			std::string *getValue() { return this->value; };
	};

	class CodeAST {
		private:
		public:
			explicit CodeAST() {}
	};
	
	class NewCodeAST : public CodeAST {
		private:
			Type *t;
			VariableAST *constructs;
			std::vector<ValueAST *> *inputs;
		public:
			explicit NewCodeAST(Type *t) : t(t), constructs(NULL), CodeAST() { this->inputs = new std::vector<ValueAST *>; }
			bool addParam(ValueAST *v);
			bool Constructs(VariableAST *v) { if(this->constructs != NULL) return false; this->constructs = v; return true; };
	};
	
	class FunctionAST {
		private:
			std::string *name;
			std::vector<TypeArrayAST *> params;
		public:
			explicit FunctionAST(std::string *name) : name(name) {}
			std::string *getName() { return this->name; }
			bool addParam(TypeArrayAST *ta);
	};
	
	class InputFunctionAST : public FunctionAST {
		public:
			explicit InputFunctionAST(std::string *name) : FunctionAST(name) {}
	};
	
	class OutputFunctionAST : public FunctionAST {
		private:
			bool hasDefault;
			bool hasNotify;
		public:
			explicit OutputFunctionAST(std::string *name) : FunctionAST(name) {}
			bool setDefault();
			bool setNotify();
	};

	class Symbol {
		private:
			std::string *ident;
			unsigned char value;
			signed long long number;
			long double fnumber;
			int line;
			int offset;
		public:
			explicit Symbol(int line, int offset) : line(line) , offset(offset) {}
			enum Type_t { Invalid , 
				EndOfFile,	// EOF
				String ,	// "my string"
				Int ,		// 1235
				Float ,		// 1.0
				Character ,	// 'c'
				Ident ,		// myident
				OpenBracket ,	// (
				CloseBracket ,	// )
				OpenSquiggle ,	// {
				CloseSquiggle ,	// }
				OpenSquare ,	// [
				CloseSquare ,	// ]
				OpenSharp , 	// <
				CloseSharp ,	// >
				QuestMark ,	// ?
				Colon ,		// :
				At ,		// @
				Comma ,		// ,
				Import ,	// import
				Class ,		// class
				Of ,		// of
				NoLock ,	// nolock
				NoRef ,		// noref
				Init ,		// init
				Input , 	// input
				Output ,	// output
				Function ,	// function
				Thread , 	// thread
				Default , 	// default
				Notify ,	// notify
				Bits ,		// bits
				SBits ,		// sbits
				FBits ,		// fbits
				Link ,		// link
				Ref ,		// ref
				UnRef , 	// unref
				Start , 	// start
				For ,		// for
				While ,		// while
				Do ,		// do
				Until ,		// until
				Break ,		// break
				Return ,	// return
				New ,		// new
				Del ,		// del
				Set ,		// set
				True ,		// true
				False ,		// false
				Inline ,	// inline
				SizeOf ,	// sizeof 
				Me		// me
				} Type;
			std::string *stringValue();
			unsigned char charValue();
			unsigned long long intValue();
			long double floatValue();
			int getLine();
			int getOffset();
			void setValue(unsigned long long value);
			void setValue(std::string *value);
			void setValue(unsigned char value);
			std::string Location();
	};
	
	class Tokenizer {
		private:
			int lastchar;
			int line;
			int offset;
			Symbol::Type_t charTypes[256];
			std::map< std::string , Symbol::Type_t > wordTypes;
			FILE *f;
			Symbol *currSymbol;
			Symbol *getSymbol(FILE *f);
		public:
			explicit Tokenizer(FILE *f) : f(f) , lastchar(' '), line(1), offset(1)
			{
				for(int i = 0; i < 256; i++) charTypes[i] = Symbol::Invalid;
			}
			Symbol *nextToken(void) { this->currSymbol = getSymbol(this->f); }
			Symbol *currToken(void) { return this->currSymbol; }
			void setSymbolChar(char c, Symbol::Type_t t) { this->charTypes[c] = t; }
			void setSymbolString(std::string word, Symbol::Type_t t) { this->wordTypes[word] = t; }
	};
	
	class Program {
		private:
			Tokenizer *tok;
			std::map<std::string, Type * > *types;
			std::map<std::string, VariableAST * > *variables;
			std::vector<CodeAST *> *code;
		public:
			Program();
			void addType(Type *t);
			Type *findType(std::string *name);
			std::map<std::string, Type *>::iterator getTypeIterBegin();
			std::map<std::string, Type *>::iterator getTypeIterEnd();
			void addOperation(CodeAST *c);
			bool addVariable(VariableAST *v);
			VariableAST *findVariable(std::string *name);
			std::map<std::string, VariableAST *>::iterator getVariableIterBegin();
			std::map<std::string, VariableAST *>::iterator getVariableIterEnd();
			
	};
	
	Program *parseFlow(FILE *f);
	bool parseImport(Tokenizer *tok, Program *p);
};

#define EXPECT(C,E,T,R) if(tok->currToken()->Type != C) { \
	cerr << endl << "Expecting " << E << " @ " << T->currToken()->Location() << endl << "Got: " << *T->currToken()->stringValue() << " of type " << T->currToken()->Type << endl; \
	return R; \
}

