#include <compiler.h>

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

using namespace LTF;

static int ASCIIvalue(int Char){
	if(Char >= '0' && Char <= '9') return (Char - '0');
	if(Char >= 'A' && Char <= 'Z') return ((Char - 'A') + 0xA);
	if(Char >= 'a' && Char <= 'z') return ((Char - 'a') + 0xA);
	return -1;
}

Symbol *Tokenizer::getSymbol(FILE *f) {
	while(!feof(f) && (isspace(lastchar)))
	{
		if(lastchar == '\n') {
			line++;
			offset = 1;
		} else if(lastchar == '\t') {
			offset += (8 - (offset & 0x7));
		} else
			offset++;
		lastchar = fgetc(f);
		if(!feof(f))
			std::cout << (char)lastchar;
	}
	
	Symbol *s = new Symbol(line, offset);
	s->Type = Symbol::Invalid;

	if(lastchar == EOF)
	{
		s->Type = Symbol::EndOfFile;
		s->setValue(new std::string("EOF"));
	}	
	else if(this->charTypes[lastchar] != Symbol::Invalid) {
		s->Type = this->charTypes[lastchar];
		char t[2] = { lastchar, '\0' };
		s->setValue(new std::string(t));
		lastchar = fgetc(f);
		std::cout << (char)lastchar;
		offset++;
	} else if(isdigit(lastchar)) { // guess its a numeric constant ... (we dont support floating point ...)
		s->Type = Symbol::Int;
		char t[30] = { lastchar, '\0' };
		int toff = 1;
		int base = 10;
		if(lastchar == '0') { // probably not base ten ...
			lastchar = fgetc(f);
			std::cout << (char)lastchar;
			t[toff] = lastchar;
			toff++;
			t[toff] = '\0';
			offset++;
			if(lastchar == 'x') { // HEX
				base = 16;
				lastchar = fgetc(f);
				t[toff] = lastchar;
				toff++;
				t[toff] = '\0';
				std::cout << (char)lastchar;
				offset++;
			} else if(lastchar == 'b') { // Binary
				base = 2;
				lastchar = fgetc(f);
				t[toff] = lastchar;
				toff++;
				t[toff] = '\0';
				std::cout << (char)lastchar;
				offset++;
			}
		}
		unsigned long long num = 0;
		while(isalnum(lastchar)) {
			int tval = ASCIIvalue(lastchar);
			if(tval >= base) {
				std::cout << "Unexpected value" << tval << "Parsing number of base " << base << std::endl;
				break; // fall out
			}
			num = (num * base) + tval;
			lastchar = fgetc(f);
			t[toff] = lastchar;
			toff++;
			t[toff] = '\0';
			std::cout << (char)lastchar;
			offset++;
		}
		if(lastchar == '.') {
			std::cerr << "Floating Point is NOT SUPPORTED" << std::endl;
		} else {
			s->setValue(new std::string(t));
			s->setValue(num);
		}
	} else if(lastchar == '\'') {
		// a quoted character
		lastchar = fgetc(f);
		s->Type = Symbol::Character;
		s->setValue((unsigned char)lastchar);
		char t[2] = { lastchar, '\0' };
		s->setValue(new std::string(t));
		lastchar = fgetc(f);
		std::cout << (char)lastchar;
		offset++;
	} else if(lastchar == '\"') {
		// a quoted string
		s->Type = Symbol::String;
		std::string str = "";
		lastchar = fgetc(f);
		std::cout << (char)lastchar;
		while(!feof(f) && lastchar != '\"') {
			str += (char)lastchar;
			lastchar = fgetc(f);
			std::cout << (char)lastchar;
			offset++;
		}
		if(feof(f)) {
			std::cerr << "EOF before \" starting @ " << s->Location() << std::endl;
			exit(-1);
		}
		s->setValue(new std::string(str));
		lastchar = fgetc(f);
		std::cout << (char)lastchar;
		offset++;
	} else {
		// assume an identifier ...
		// hopefully no one does anything funky ...
		std::string str = "";
		while(!isspace(lastchar) && charTypes[lastchar] == Symbol::Invalid) {
			str += (char)lastchar;
			lastchar = fgetc(f);
			std::cout << (char)lastchar;
			offset++;
		}
		// check for known keywords
		s->Type = wordTypes[str];
		if(s->Type == Symbol::Invalid)
			s->Type = Symbol::Ident;
		s->setValue(new std::string(str));
	}
	// eat the last char
	
	return s;
}
